;*******************************************************************************
;
;	Filename:		MSMV-3.asm
;	Date:				15 Jun 2016
;	File Version:	1.0
;	Author:			Larry Cicchinelli
;	Description:
;		This program uses a PIC 12F1572 to emulate a 555 timer in its
;				Mono-Stable MultiVibrator mode
;
; Both starting and terminating the pulse are done via interrupts so are
; independant of main program execution
;
;IMPORTANT - any references to time assume that Fosc = 32MHz
;
;		The A/D is used to read the voltage on a potentiometer.  That reading
;		is used to determine the desired delay time.  However,
;		the range is fixed and is determined by macro definitions.
;		Timer 1 is used for timing the pulse width.  See Timer 1 macros.
;
; The pulse width is determined by the following:
;	CPU clock, Timer 1 source, timer 1 prescale, A/D reading and A/D multiplier
; For example, if clock speed = 32MHz and T1_SOURCE = Fosc/4 then
;	PW(us) = (4 * T1_PRESCALE * A/D reading * A/D multiplier) / 32
;			 = (T1_PRESCALE * A/D reading * A/D multiplier) / 8
; Macros T1_SOURCE, T1_PRESCALE and AD_MULT are all used to select the basic
; resolution of the pulse width.  Using the default values (T1SRC_FOSC4,
;	T1PRESCALE_1 and ADMULT_1) the time resolution is 125ns.  The widest pulse
;	is 125ns * 1023 = 127.875us + (15 to 17) instruction cycles
;
; The delay between the trigger and the output pulse is:
;		3 to 5 instruction cycles due to interrupt latency
;		+ 12 instruction cycles in the ISR (if retrigger is disabled)
;
; The A/D sample clock is CPUclk/32 = 1MHz.  The total time to get an A/D
;	reading is about 15us.
;
; Options:
;		ENABLE_RETRIGGER
;			= 1 = enable retriggering
;			= 0 = no retrigger (default)
;		EDGE_REG
;			= IOCAP = trigger on positive edge of input
;			= IOCAN = trigger on negative edge of input
;		T1_SOURCE - macro to define the Timer 1 source
;			= T1SRC_FOSC4 - use Fosc/4									(default)
;			= T1SRC_FOSC - use Fosc (32MHz)
;			= T1SRC_LFINTOSC - use the Low Frequency internal oscillator (~31KHz)
;		T1_PRESCALE - macro to define the Timer 1 prescale value
;			= T1PRESCALE_1 - divide source by 1
;			= T1PRESCALE_2 - divide source by 2
;			= T1PRESCALE_4 - divide source by 4
;			= T1PRESCALE_8 - divide source by 8						(default)
;		AD_MULT - macro to multiply the A/D reading
;			= ADMULT_1 = multiply by 1									(default)
;			= ADMULT_2 = multiply by 2
;			= ADMULT_4 = multiply by 4
;			= ADMULT_8 = multiply by 8
;			= ADMULT_16 = multiply by 16
;			= ADMULT_32 = multiply by 32
;			= ADMULT_64 = multiply by 64
;
; Using the default values for Timer 1, its clock will be 1Mhz giving a
; resolution of 1us.
;
;		PIC12F1572 I/O:
;			pin 1: VDD
;			pin 2: RA5							Pulse output
;			pin 3: RA4, AN3					Trigger
;			pin 4: ~MCLR
;			pin 5: RA2, AN2					Pulse width adjust
;			pin 6: ICSPCLK
;			pin 7: ISCPDAT
;			pin 8: VSS
;
;*******************************************************************************
;*******************************************************************************

		#include p12F1572.inc

; CONFIG1
; __config 0xF9E4
 __CONFIG _CONFIG1, _FOSC_INTOSC & _WDTE_OFF & _PWRTE_OFF & _MCLRE_ON & _CP_OFF & _BOREN_OFF & _CLKOUTEN_OFF
; CONFIG2
; __config 0xFEFF
 __CONFIG _CONFIG2, _WRT_OFF & _PLLEN_OFF & _STVREN_ON & _BORV_LO & _LPBOREN_OFF & _LVP_ON


;*******************************************************************************
; Definitions used by the program which are associated with specific I/O pins:
;*******************************************************************************
AV_PW			equ	RA2					; pin for analog voltage for the pulse width
I_TRIGGER	equ	RA4					; input trigger pin
O_PULSE		equ	RA5					; output pulse pin

ANSEL_VAL	equ	(1<<AV_PW)			; initialize ANSELA
ADCH_PW		equ	(AV_PW<<CHS0) | (1<<ADON ) | (1<<ADGO) ; ADCON0 value
TRISA_VAL	equ	~(1<<O_PULSE)		; initialize TRISA
	
;===============================================================================
; user options
;===============================================================================

ENABLE_RETRIGGER equ 0					; 0 = no re-trigger (default)
												; 1 = enable re-trigger

EDGE_REG equ	IOCAP						; IOCAP = trigger on positive edge (default)
												; IOCAN = trigger on negative edge

; ========= Timer 1 definitions ==========
; IMPORTANT: see the notes immediately preceding the ISR.
;	When using T1SRC_FOSC4, you should set T1_PRESCALE to
;	T1PRESCALE_8 due to instruction execution time.  This yields 1us
;	resolution.
;	One of the allowable Timer clock inputs is the 32MHz clock.  Using this is
;	a problem, again due to istruction execution time.
;	CPU clock source
;
T1SRC_FOSC4		equ	0					; Fosc/4 (default)
T1SRC_LFINTOSC	equ	3					; Low Frequency Internal Oscillator (~31KHz)

T1_SOURCE	equ	T1SRC_FOSC4			; use one of the T1SRC_...values

; Timer 1 prescale
T1PRESCALE_1 equ	0
T1PRESCALE_2 equ	1
T1PRESCALE_4 equ	2
T1PRESCALE_8 equ	3						; (default)

T1_PRESCALE	equ	T1PRESCALE_8		; use one of the T1PRESCALE_...values

;; ========= A/D multiplier definitions ==========
; The A/D reading will be shifted left the requested nbr of times in order to
; increase the range of the PW - which also reduces the resolution.
ADMULT_1			equ	0					; (default)
ADMULT_2			equ	1
ADMULT_4			equ	2
ADMULT_8			equ	3
ADMULT_16		equ	4
ADMULT_32		equ	5
ADMULT_64		equ	6

AD_MULT			equ	ADMULT_1			; use one of the ADMULT_ values


;*******************************************************************************
; Common RAM allocations: 16 bytes
;*******************************************************************************
Access_RAM		udata_shr	0x70		; 16 bytes starting at address 0x70
; IMPORTANT: T1Value MUST start on an even byte boundary
T1Value			res	4					; double buffer for Timer 1 preset value
ADvalue			res	2					; two bytes for A/D reading
TestValue		res	2
			
;*******************************************************************************
; Interrupt Vectors
;*******************************************************************************

	org	0
	goto	Main

; The ISR handles interrupts from both Timer1 and Interrupt On Change
; The IOC signal is used to trigger the output pulse
; The Timer interrupt is used to turn off the output pulse
;
; note: Interrupt latency is 3 to 5 instruction cycles
; The Timer ISR takes 16 instruction cycles +3 or 4 for interrupt latency
; The IOC ISR takes 13 instruction cycles +3 o 5 for interrupt latency
; Therefore, the minimum pulse width will be about 32 instruction cycles.
; With the default definitions, this is equivalent to 4.25us

	org	0x04								; interrupt vector
	BANKSEL	0								; PORTA, TIMER 1, & PIRx registers in bank 0
	btfss		PIR1, TMR1IF				; skip next if Timer 1 interrupt
	goto		ISR.20						; must be IOC
; handle Timer 1 overflow
	bcf		PIR1, TMR1IF				; clear interrupt flag
	bcf		T1CON, TMR1ON				; disable Timer 1
	bcf		IOCAF, I_TRIGGER			; clear interrupt flag (just in case)
	BANKSEL	LATA
	bcf		LATA, O_PULSE				; turn off pulse
; insert Timer 1 preset value
	BANKSEL	0								; PORTA, TIMER 1, & PIRx registers in bank 0
	moviw		0[FSR1]							; get low byte of Timer 1 load value
	movwf		TMR1L							; store it
	moviw		1[FSR1]							; get high byte, point back to low byte
	movwf		TMR1H							; store it
	bsf		INTCON, IOCIE				; enable Interrupt On Change
	retfie

ISR.20										; handle IOC interrupt
	BANKSEL	IOCAF
;	btfss		IOCAF, I_TRIGGER			; skip next if trigger detected
;	retfie
; handle trigger
	bcf		IOCAF, I_TRIGGER			; clear interrupt flag
#if ENABLE_RETRIGGER == 0		; if retrigger disabled
	bcf		INTCON, IOCIE				; disable Interrupt On Change
#endif
	BANKSEL	0								; PORTA, TIMER 1, & PIRx registers in bank 0
	bsf		T1CON, TMR1ON				; enable Timer 1
	BANKSEL	LATA
	bsf		LATA, O_PULSE				; pulse on
	retfie
; Interrupt vectors

;*******************************************************************************
; MAIN PROGRAM
;*******************************************************************************

MAIN_PROG CODE                      ; let linker place main program

Main
	call		init							; initialize the system hardware
; load INDF1 with address of T1Value - used by ISR to load Timer 1 registers
	movlw		T1Value
	movwf		FSR1L
	clrf		FSR1H
;
	call		AD.Read						; initial read and setup of double buffer
	bsf		INTCON, IOCIE				; enable Interrupt On Change
	
Main.10
	call		AD.Read						; read A/D
;
; your program goes here
;
	goto		Main.10
; Main
;*******************************************************************************


AD.Read
; read the 10 bit A/D
; exit with:
;		10 bit value in ADvalue and ADvalue+1, shifted by AD_MULT and negated
	BANKSEL	ADCON0
	movlw		ADCH_PW
	movwf		ADCON0						; set the channel, A/D on and triggered
AD.Read.10
	btfsc		ADCON0, ADGO				; skip next if done conversion
	goto		AD.Read.10					; wait for done

	movf		ADRESL, w					; get the low byte
	movwf		ADvalue						; save it
	movf		ADRESH, w					; get the high byte
	andlw		3								; ensure 10 bit value
	movwf		ADvalue+1					; save it

;;;;;; load a dummy for testing
;	movfw		TestValue
;	movwf		ADvalue
;	movfw		TestValue+1
;	movwf		ADvalue+1
;;;;;;

; test for 0, if value=0 then make ADvalue = 1
	iorwf		ADvalue, w					; w = 0 if ADRESL and ADRESH are both 0
	btfsc		STATUS, Z					; skip next if w <> 0
	incf		ADvalue, f					; make value 1

#if AD_MULT > 0
; adjust A/D reading based on AD_MULT
	movlw		AD_MULT						; get the multiplier
AD.Read.20
	lslf		ADvalue, f					; shift low byte: C=bit7, bit0=0
	rlf		ADvalue+1, f				;		and high byte: bit0=C
	decfsz	WREG, f						; update shift counter, skip next if done
	goto		AD.Read.20
#endif

; negate the reading by complementing and incrementing
	comf		ADvalue, w					; process low byte
	addlw		1								; may set C bit
	movwf		ADvalue						; store it
	comf		ADvalue+1, f				; process high byte
	btfsc		STATUS, C					; skip next if no carry
	incf		ADvalue+1, f

; now store the reading into the timer preset memory
	btfsc		FSR0L, 1						; skip next if pointing to buffer 0
	goto		AD.Read.30
; here if ISR is using buffer 0 - put value into buffer 1
	movfw		ADvalue						; copy 
	movwf		T1Value+2					;		reading
	movfw		ADvalue+1					;			to
	movwf		T1Value+3					;				Timer 1 preset
	bsf		FSR0L, 1						; tell ISR to use buffer 1
	BANKSEL	0								; PORTA, TIMER 1, & PIRx registers in bank 0
	return

; here if ISR is using buffer 1 - put value into buffer 0
AD.Read.30
	movfw		ADvalue						; copy 
	movwf		T1Value+0					;		reading
	movfw		ADvalue+1					;			to
	movwf		T1Value+1					;				Timer 1 preset
	bcf		FSR0L, 1						; tell ISR to use buffer 0
	BANKSEL	0								; PORTA, TIMER 1, & PIRx registers in bank 0
	return
; AD.Read


;===============================================================================
;									Initialize Hardware
;===============================================================================

init	; initialize the system hardware

; Oscillator - set to 32MHz using 8MHz clock and PLL
	BANKSEL	OSCCON
	movlw		(1<<SPLLEN) | (b'1110'<<IRCF0) | (b'00'<<SCS0)
	movwf		OSCCON						; enable PLL, IntOsc=8Mhz, clock determined by CONFIG
												; *** SCS must be '00' to allow PLL operation ***
init.10		; wait for PLL ready
	btfss		OSCSTAT, PLLR				; skip next if PLL ready
	goto		init.10

; Port A
	BANKSEL	PORTA
	clrf		PORTA
	BANKSEL	TRISA
	movlw		TRISA_VAL
	movwf		TRISA
	BANKSEL	ANSELA
	movlw		ANSEL_VAL
	movwf		ANSELA

; A/D convertor
	BANKSEL	ADCON0
	movlw		(1<<ADON)					; turn on A/D
	movwf		ADCON0
	movlw		(1<<ADFM) | (b'010'<<ADCS0) | (0<<ADPREF0)
	movwf		ADCON1						; right justify data, clk=Fosc/32, VRPOS=VDD
	clrf		ADCON2						; use ADCON0(1<<ADGO) as trigger

; Timer 1
	BANKSEL	T1CON
	movlw		(T1_SOURCE<<TMR1CS0) | (T1_PRESCALE<<T1CKPS0) | (1<<NOT_T1SYNC)
	movwf		T1CON							; set source and prescale, Timer off
	clrf		T1GCON						; do not use gate

; Interrupts
	movlw		(1<<GIE) | (1<<PEIE)		; enable global and peripheral interrupts
	movwf		INTCON
	BANKSEL	PIE1
	movlw		(1<<TMR1IE)					; enable Timer 1 interrupt
	movwf		PIE1
	BANKSEL	IOCAP
	bsf		EDGE_REG, I_TRIGGER		; set the IOC polarity

	BANKSEL	0								; PORTA, TIMER 1, & PIRx registers in bank 0
	return
; init

	END
